"""Command-line interface."""
import configparser
import os
from difflib import get_close_matches
from pathlib import Path
from typing import List
from typing import Optional
from typing import Set

import typer
from tabulate import tabulate

from loglicense import LicenseLogger
from loglicense.utils import DependencyFileParser


app = typer.Typer()
OK, ERR, FAIL_UNDER = typer.Exit(code=0), typer.Exit(code=1), typer.Exit(code=2)


def search_dependency_file() -> str:
    """Searches for supported files in current directory.

    Returns:
        str: First supported file found in directory

    Raises:
        Exception: Fails if no supported files found

    """
    supported_files = DependencyFileParser().parsers.keys()
    files = os.listdir(".")
    found_files = [x for x in files if x.lower() in supported_files]
    if len(found_files) == 0:
        raise Exception("No supported files found in current directory.")
    dependency_file = found_files[0]
    return dependency_file


@app.command()
def report(
    dependency_file: Optional[str] = None,
    package_manager: str = "pypi",
    info_columns: Optional[List[str]] = None,
    tablefmt: str = "pipe",
    develop: bool = False,
    output_file: Optional[str] = None,
) -> None:
    """Document licenses of packages in dependency file.

    Args:
        dependency_file: Specify file to crawl dependencies for.
            Defaults to search directory for supported files.
        package_manager: Which type of package manager to evaluate.
            Defaults to pypi for python.
        info_columns: Information to include in table to log
        tablefmt: Tabulates formatting argument
        develop: Whether to include development dependencies
        output_file: File to save table of licenses in
    """
    info_columns = info_columns if info_columns else ["name", "license"]

    if not dependency_file:
        dependency_file = search_dependency_file()

    license_log = LicenseLogger(
        dependency_file=dependency_file,
        package_manager=package_manager,
        info_columns=info_columns,
        develop=develop,
    )

    license_table = tabulate(
        license_log.log_licenses(), tablefmt=tablefmt, headers="firstrow"
    )

    if output_file:
        output_filepath = Path(output_file)
        output_filepath.touch(exist_ok=True)
        output_filepath.write_text(license_table)
    else:
        print(license_table)


@app.command()
def check(
    dependency_file: str,
    config_file: str = ".loglicense",
    package_manager: str = "pypi",
    develop: bool = False,
    show_report: bool = True,
) -> None:
    """Check licenses of packages in dependency file.

    Args:
        dependency_file: File to crawl dependencies for
        config_file: Config for parameters of the license check
        package_manager: Which type of package manager to evaluate.
            Defaults to pypi for python.
        develop: Whether to include development dependencies
        show_report: Print information regarding licences checked

    Raises:
        OK: 0 exit code
        ERR: 1 exit code
        FAIL_UNDER: 2 exit code
    """
    cf = configparser.ConfigParser()
    cf.read(config_file)
    config = cf["loglicense"]

    if not dependency_file:
        dependency_file = search_dependency_file()

    license_log = LicenseLogger(
        dependency_file=dependency_file,
        package_manager=package_manager,
        info_columns=["name", "license"],
        develop=develop,
    )

    allowed = {
        x.lower().strip() for x in config.get("allowed", "").split(",") if x.strip()
    }
    banned = {
        x.lower().strip() for x in config.get("banned", "").split(",") if x.strip()
    }
    validated = {
        x.lower().strip() for x in config.get("validated", "").split(",") if x.strip()
    }
    results = validate_requirements(license_log, allowed, banned, validated)

    if show_report:
        print(f"Found {len(results)-1} dependencies")
        pretty_print = tabulate(results, tablefmt="pipe", headers="firstrow")
        print(pretty_print)

    result_status = [x[-1] for x in results[1:]]

    if any([x == "Banned" for x in result_status]):
        raise ERR

    try:
        license_coverage = int(
            (result_status.count("Allowed") / len(result_status)) * 100
        )
    except ZeroDivisionError:
        license_coverage = 100

    if "coverage" in config:
        target_cov = int(config.get("coverage"))
        print(
            f"Target license coverage ({target_cov}%) "
            f"and actual coverage: {license_coverage}%"
        )
        if license_coverage < target_cov:
            raise FAIL_UNDER

    raise OK


def validate_requirements(
    license_logger: LicenseLogger,
    allowed: Set[str],
    banned: Set[str],
    validated: Set[str],
) -> List[List[str]]:
    """Compare licenses to requirements.

    Args:
        license_logger: File to crawl dependencies for
        allowed: Whether to include development dependencies
        banned: Print information regarding licences checked
        validated: Manually validated packages

    Returns:
        List[List[str]]: Returns results of validation of license
    """
    license_log = license_logger.log_licenses()
    results = [license_log[0] + ["Status"]]
    for lib in license_log[1:]:
        # handle multiple licenses
        for lib_license in lib[-1].split("\n"):
            matches = None
            if lib[0] in validated:
                results.append(
                    [lib[0], lib[-1].replace("\n", ", "), "Manually validated"]
                )
                continue
            if banned:
                matches = get_close_matches(
                    lib_license.lower().replace("license", ""), banned
                )
                if matches:
                    results.append([lib[0], lib_license, "Banned"])
                    continue

            if allowed:
                matches = get_close_matches(
                    lib_license.lower().replace("license", ""), allowed
                )
                if matches:
                    results.append([lib[0], lib_license, "Allowed"])

                if not matches:
                    results.append([lib[0], lib_license, "Unknown"])

            else:
                if not matches:
                    results.append([lib[0], lib_license, "Allowed"])

    return results


if __name__ == "__main__":
    app()



"""Utility functions for loglicense module."""
from pathlib import Path
from typing import List
from typing import Set

import toml


class DependencyFileParser:
    """Main module for DependencyFileParser."""

    def __init__(self) -> None:
        super().__init__()
        self.method_prefix = "parse_"
        self.parsers = {
            self.__get_dependency_filename(attribute): getattr(self, attribute)
            for attribute in dir(self)
            if callable(getattr(self, attribute))
            and attribute.startswith("parse") is True
        }

    def __get_dependency_filename(self, attribute: str) -> str:
        """Format the parser function name into original filename.

        Args:
            attribute: Name of the parser function

        Returns:
            str: Filename of the dependency file
        """
        dependency_filename = attribute.replace(self.method_prefix, "").replace(
            "_", "."
        )
        return dependency_filename

    @staticmethod
    def parse_poetry_lock(license_path: Path, develop: bool = False) -> List[str]:
        """Main module for DependencyFileParser.

        Args:
            license_path: Path to license file (poetry.lock)
            develop: Whether to include development dependencies

        Returns:
            List[str]: List of the names of python depedencies in lock file
        """
        output: List[str] = []
        included_categories: Set[str] = {"main"}

        if develop:
            included_categories.add("dev")

        license_file = toml.load(license_path)
        for pkg in license_file.get("package", []):
            if pkg["category"] in included_categories:
                output.append(pkg.get("name", ""))
        return output



"""LogLicence main module."""
import json
import logging
from pathlib import Path
from typing import Any
from typing import List
from typing import Optional
from urllib.request import urlopen

from loglicense.utils import DependencyFileParser


logger = logging.getLogger("licenselogger")


class LicenseLogger:
    """Main module for logging licenses.

    Args:
        dependency_file: File to crawl dependencies for
        package_manager: Which type of package manager to evaluate.
            Defaults to pypi for python.
        info_columns: Information to include in table to log
        develop: Whether to include development dependencies

    """

    def __init__(
        self,
        dependency_file: str,
        package_manager: str = "pypi",
        info_columns: Optional[List[str]] = None,
        develop: bool = False,
    ):
        super().__init__()
        self.dependency_file = Path(dependency_file)
        self.package_manager = package_manager
        self.info_columns = info_columns if info_columns else ["name", "license"]
        self._parser_args = {"develop": develop}

        if not self.dependency_file.is_file():
            raise ValueError("Path must be a file")

        self.parser = DependencyFileParser().parsers[self.dependency_file.name]

        if self.package_manager == "pypi":
            self.library_url = "https://pypi.python.org/pypi/XXX/json"
        # elif self.package_manager == "npm":
        #     self.library_url = "https://registry.npmjs.org/XXX/latest"
        else:
            raise NotImplementedError("Only supports pypi dependencies")

    def log_licenses(
        self,
    ) -> List[List[str]]:
        """Fetches license package metadata for the given dependency file.

        Returns:
            List[List[str]]: Metadata from licenses found in dependency
            file.
        """
        self.licenselog_ = [[x.capitalize() for x in self.info_columns]]

        for libname in self.parser(self.dependency_file, **self._parser_args):
            pkg_metadata = self.get_license_metadata(libname)
            lib_metadata = []
            if not pkg_metadata:
                lib_metadata.append(libname)
                lib_metadata.extend(
                    ["Not found" for x in range(len(self.info_columns) - 1)]
                )
            else:
                for col in self.info_columns:
                    licenses = pkg_metadata.get(col, "")

                    if not licenses:
                        licenses = ""
                    if col == "license" and licenses.strip() == "":
                        classifiers = pkg_metadata.get("classifiers", "")
                        licenses = [
                            classifier.replace("License :: ", "").replace(
                                "OSI Approved :: ", ""
                            )
                            for classifier in classifiers
                            if classifier.startswith("License")
                        ]
                        licenses = "\n".join(licenses)

                    lib_metadata.append(licenses)

            self.licenselog_.append(lib_metadata)

        return self.licenselog_

    def get_license_metadata(self, libname: str) -> Any:
        """Fetch information from package manager site.

        Args:
            libname: Name of the package to fetch information regarding

        Returns:
            Any: The metadata of the library
        """
        lib_url = self.library_url.replace("XXX", libname)
        try:
            with urlopen(lib_url) as libfile:
                output = json.loads(libfile.read().decode())

                if self.package_manager == "pypi":
                    output = output.get("info", {})
            return output

        except Exception:
            logger.warning(f"{libname}: error in fetching metadata")
            return None

    def is_logged(self) -> bool:
        """Check if logged.

        Returns:
            bool: Whether logged or not
        """
        logged = [v for v in dir(self) if v.endswith("_") and not v.startswith("__")]

        return bool(len(logged))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['loglicense']

package_data = \
{'': ['*']}

install_requires = \
['pathlib>=1.0.1,<2.0.0',
 'tabulate>=0.8.10,<0.9.0',
 'toml>=0.10.2,<0.11.0',
 'typer>=0.6.1,<0.7.0',
 'types-tabulate>=0.8.11,<0.9.0',
 'types-toml>=0.10.8,<0.11.0']

entry_points = \
{'console_scripts': ['loglicense = loglicense.__main__:app']}

setup_kwargs = {
    'name': 'loglicense',
    'version': '0.1.0',
    'description': 'Log License',
    'long_description': "# Log License\n\n[![PyPI](https://img.shields.io/pypi/v/loglicense.svg)][pypi_]\n![Downloads](https://img.shields.io/pypi/dm/loglicense)\n[![Status](https://img.shields.io/pypi/status/loglicense.svg)][status]\n[![Python Version](https://img.shields.io/pypi/pyversions/loglicense)][python version]\n[![License](https://img.shields.io/pypi/l/loglicense)][license]\n\n[![Read the documentation at https://loglicense.readthedocs.io/](https://img.shields.io/readthedocs/loglicense/latest.svg?label=Read%20the%20Docs)][read the docs]\n[![Tests](https://github.com/martincjespersen/loglicense/workflows/Tests/badge.svg)][tests]\n[![Codecov](https://codecov.io/gh/martincjespersen/loglicense/branch/main/graph/badge.svg)][codecov]\n\n[![pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)][pre-commit]\n[![Black](https://img.shields.io/badge/code%20style-black-000000.svg)][black]\n\n[pypi_]: https://pypi.org/project/loglicense/\n[status]: https://pypi.org/project/loglicense/\n[python version]: https://pypi.org/project/loglicense\n[read the docs]: https://loglicense.readthedocs.io/\n[tests]: https://github.com/martincjespersen/loglicense/actions?workflow=Tests\n[codecov]: https://app.codecov.io/gh/martincjespersen/loglicense\n[pre-commit]: https://github.com/pre-commit/pre-commit\n[black]: https://github.com/psf/black\n\n## Features\n\n- Report and save log of licenses included in project\n\n### Supported dependency files\n\n- poetry.lock\n\n### Supported package managers\n\n- pypi\n\n## Installation\n\nYou can install _Log License_ via [pip] from [PyPI]:\n\n```console\n$ pip install loglicense\n```\n\nor using [Poetry]\n\n```console\n$ poetry add loglicense\n```\n\n## Quick example\n\nPlease see the [Command-line Reference] for details.\n\n```console\n$ loglicense report path_to/poetry.lock\n```\n\nExample output of this project's poetry.lock file:\n\n```console\n| name               | license                            |\n|:-------------------|:-----------------------------------|\n| click              | BSD-3-Clause                       |\n| colorama           | BSD                                |\n| importlib-metadata | Apache Software License            |\n| pathlib            | MIT License                        |\n| tabulate           | MIT                                |\n| toml               | MIT                                |\n| typer              | MIT License                        |\n| typing-extensions  | Python Software Foundation License |\n| zipp               | MIT License                        |\n```\n\nAlternatively you can let it search the executed directory for any supported file\n\n```console\n$ loglicense report\n```\n\n## Features to implement\n\n- Support npmjs package manager (and package.json/package-lock.json)\n- Support Pipfile, pyproject.toml, Pipfile.lock, requirements.txt\n\n## Contributing\n\nContributions are very welcome.\nTo learn more, see the [Contributor Guide].\n\n## License\n\nDistributed under the terms of the [Apache 2.0 license][license],\n_Log License_ is free and open source software.\n\n## Issues\n\nIf you encounter any problems,\nplease [file an issue] along with a detailed description.\n\n## Credits\n\nThis project was generated from [@cjolowicz]'s [Hypermodern Python Cookiecutter] template.\n\n[@cjolowicz]: https://github.com/cjolowicz\n[pypi]: https://pypi.org/\n[hypermodern python cookiecutter]: https://github.com/cjolowicz/cookiecutter-hypermodern-python\n[file an issue]: https://github.com/martincjespersen/loglicense/issues\n[pip]: https://pip.pypa.io/\n\nThis project is greatly inspired by [dep-license] created by [Abdulelah Bin Mahfoodh].\n\n[dep-license]: https://github.com/abduhbm/dep-license\n[abdulelah bin mahfoodh]: https://github.com/abduhbm\n[poetry]: https://python-poetry.org/\n\n<!-- github-only -->\n\n[license]: https://github.com/martincjespersen/loglicense/blob/main/LICENSE\n[contributor guide]: https://github.com/martincjespersen/loglicense/blob/main/CONTRIBUTING.md\n[command-line reference]: https://loglicense.readthedocs.io/en/latest/usage.html\n",
    'author': 'Martin Closter Jespersen',
    'author_email': 'martincjespersen@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/martincjespersen/loglicense',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
